# comtypes.client
"""high level client level COM support module."""

# Nearly done (just have to doc it in the docstrings):
# - provide a way to specify a friendly typelib name in CreateObject
#   (for py2exe, and for the programmer)
# - generate the friendly_name module in comtypes.gen
# - generate the friendly_name module even if the unfriendly module
#   already exists
# Must rename make_module and _make_module!!!

################################################################
#
# TODO: 
#
# - rename make_module into _make_module, on other words make it private
# - rename wrap
#
# - think about constants in typelib wrappers
#
# - a GetObject function?
#
# - make CreateObject also accept a GUID (string and instance?)
#
# - beautify the code generator output (import statements at the top)
#
################################################################

import sys, os, new
from ctypes import *

from comtypes import GUID, CoCreateInstance
from comtypes import CLSCTX_INPROC_SERVER, CLSCTX_LOCAL_SERVER, CLSCTX_SERVER
from comtypes.automation import IDispatch
from comtypes.typeinfo import IProvideClassInfo
import comtypes.gen

# determine the place where generated modules live

# Fix this to recreate the comtypes\gen\__init__.py file if it was
# accidentially deleted!
gen_dir = comtypes.gen.__path__[0]
if not os.path.isdir(gen_dir):
    gen_dir = None
### for testing
##gen_dir = None
    
def _my_import(fullname):
    # helper function to import dotted modules
    mod = __import__(fullname)
    for name in fullname.split(".")[1:]:
        mod = getattr(mod, name)
    return mod

def _name_module(tlib):
    libattr = tlib.GetLibAttr()
    modname = "_%s_%s_%s_%s" % \
              (str(libattr.guid)[1:-1].replace("-", "_"),
               libattr.lcid,
               libattr.wMajorVerNum,
               libattr.wMinorVerNum)
    return "comtypes.gen." + modname

FORCE = False

def make_module(tlib, friendly_name=None):
    """Create a module wrapping a COM typelibrary on demand.

    'tlib' must be an ITypeLib COM pointer instance.  'friendly_name'
    specifies the directory plus basename of a module importing
    everything from the typelib wrapper.

    Determines the module name from the typelib attributes, then tries
    to import.  If that fails, the module is regenerated in the
    comtypes.gen package.

    If the comtypes.gen __path__ is not a directory (in a frozen
    executable it lives in a zip archive), the module is only created
    in memory.

    Example:

        make_module(LoadTypeLibEx("shdocvw.dll"),
                    friendly_name="IE_typelib")

    would create a module named
       comtypes.gen._EAB22AC0_30C1_11CF_A7EB_0000C05BAE0B_0_1_1
    containing the wrapper code for the type library used by InternetExplorer.
    The 'comtypes.gen.IE_typelib' module is created as an alias.
    """
    fullname = _name_module(tlib)
    mod = _make_module(tlib, fullname)
    if friendly_name is None:
        return mod
    try:
        return _my_import("comtypes.gen." + friendly_name)
    except:
        modname = fullname.split(".")[-1]
        code = "from comtypes.gen import %s\nglobals().update(%s.__dict__)\n" % (modname, modname)
        if gen_dir is None:
            mod = new.module("comtypes.gen." + friendly_name)
            exec code in mod.__dict__
            sys.modules[mod.__name__] = mod
            setattr(comtypes.gen, friendly_name, mod)
            return mod
        # create in file system, and import it
        ofi = open(os.path.join(gen_dir, friendly_name + ".py"), "w")
        ofi.write(code)
        ofi.close()
        return _my_import("comtypes.gen." + friendly_name)
        
def _make_module(tlib, fullname):
    # helper which creates and imports the real typelib wrapper module.
    modname = fullname.split(".")[-1]
    try:
        mod = _my_import(fullname)
    except Exception:
        # We generate the module on ANY error we get.
        from comtypes.tools.tlbparser import generate_module
        import time
        start = time.clock()
        if gen_dir is None:
            import cStringIO
            ofi = cStringIO.StringIO()
        else:
            ofi = open(os.path.join(gen_dir, modname + ".py"), "w")
        # use warnings.warn, maybe?
        print "# Generating", modname
        generate_module(tlib, ofi, make_module, _name_module)
##        print "# generating took %s seconds" % (time.clock() - start)

        if gen_dir is None:
            code = ofi.getvalue()
            mod = new.module(fullname)
            exec code in mod.__dict__
            sys.modules[fullname] = mod
            setattr(comtypes.gen, modname, mod)
        else:
            ofi.close()
            mod = _my_import(fullname)
            reload(mod)
        return mod
    else:
        # first import above succeeded.
        return mod

# XXX rename this!
def wrap(punk, friendly_name=None):
    """Try to QueryInterface a COM pointer to the 'most useful'
    interface.
    
    Get type information for the provided object, either via
    IDispatch.GetTypeInfo(), or via IProvideClassInfo.GetClassInfo().
    Generate a wrapper module for the typelib, and QI for the
    interface found.
    """
    if not punk: # NULL COM pointer
        return punk # or should we return None?
    # find the typelib and the interface name
    try:
        pci = punk.QueryInterface(IProvideClassInfo)
        tinfo = pci.GetClassInfo() # TypeInfo for the CoClass
        # find the interface marked as default
        for index in range(tinfo.GetTypeAttr().cImplTypes):
            if tinfo.GetImplTypeFlags(index) == 1:
                break
        else:
            # should we simply use the first interface now?
            raise TypeError, "No default interface found"
        href = tinfo.GetRefTypeOfImplType(index)
        tinfo = tinfo.GetRefTypeInfo(href)
    except WindowsError: # XXX we need a COMError!
        try:
            pdisp = punk.QueryInterface(IDispatch)
            # TypeInfo for the interface
            tinfo = pdisp.GetTypeInfo()
        except WindowsError:
            return punk
    itf_name = tinfo.GetDocumentation(-1)[0] # interface name
    tlib = tinfo.GetContainingTypeLib()[0] # typelib

    # import the wrapper, generating it on demand
    mod = make_module(tlib, friendly_name)
    # Python interface class
    interface = getattr(mod, itf_name)
    # QI for this interface
    result = punk.QueryInterface(interface)
    # Hm, this makes the constants in the module (but not only the
    # constants) available to the returned object.
    result._constants = mod
    return result

# Should we do this for POINTER(IUnknown) also?
POINTER(IDispatch).__ctypes_from_outparam__ = wrap

def CreateObject(progid,
                 clsctx=CLSCTX_SERVER,
                 friendly_name=None):
    """Create a COM object from 'progid', and try to QueryInterface()
    it to the most useful interface, generating typelib support on
    demand.  A pointer to this interface is returned.

    'progid' may be a string like "InternetExplorer.Application",
       a string specifying a clsid, a GUID instance, or an object with
       a _clsid_ attribute.
    'clsctx' specifies how to create the object.
    'friendly_name' allows to specify a readable alias name for the typelib wrapper.
    """
    if hasattr(progid, '_clsid_'):
        progid = progid._clsid_

    if isinstance(progid, GUID):
        clsid = progid
    elif isinstance(progid, basestring):
        if progid.startswith("{"):
            clsid = GUID(progid)
        else:
            clsid = GUID.from_progid(progid)
    else:
        raise TypeError, progid
    punk = CoCreateInstance(clsid, clsctx=clsctx)
    return wrap(punk, friendly_name)

################################################################
#
# Hm, don't these belong into comtypes\__init__.py?  And, if we derive
# this from c_wchar_p, we could even avoid the __ctypes_from_outparm__
# method, because that already returns a unicode string.
class TaskMemString(POINTER(c_wchar)):
    """A zero terminated unicode string, freed by CoTaskMemFree after
    use.
    """

    # When retrieved from an 'out' parameter, the
    #__ctypes_from_outparam__ method is called.  Return a Python
    #unicode string, instead of the LP_c_wchar instance itself.
    #
    _type_ = c_wchar # the type pointed to

    def __ctypes_from_outparam__(self):
        return wstring_at(self)

    def __del__(self, _free=windll.ole32.CoTaskMemFree):
        _free(self)

# code generation could probably generate commented out code (or not commented out?)
# with "magic_replace(POINTER(c_wchar), TaskMemString)" and a comment about that.
def magic_replace(old, new, debug=False):
    """This function is used to replace POINTER types in generated
    code.  It must be called in the module where the replacement
    should take place, BEFORE the generated code is executed.  The
    ctypes' POINTER symbol must be available before the call, this
    function replaces the POINTER function with a specialized one in
    the calling module.
    
        from ctypes import *
        POINTER = magic_replace(POINTER(c_wchar), TaskMemString)
    
    After that, POINTER is modified to return POINTER(TaskMemString)
    whenever POINTER(POINTER(c_wchar)) is called.
    Uses sys._getframe to change the globals in the calling module.
    """
    frame = sys._getframe(1)

    POINTER = frame.f_globals["POINTER"]

    def hack_POINTER(typ):
        if typ == old:
##            if debug:
##                print "returns %s instead of %s" % (POINTER(new), POINTER(typ))
            return POINTER(new)
        return POINTER(typ)

    frame.f_globals["POINTER"] = hack_POINTER

    return hack_POINTER

__all__ = ["CreateObject", "magic_replace", "make_module", "wrap", "TaskMemString"]
